@@ -1,11 +1,11 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/child.c,v 1.1.1.1 2006/09/08 03:36:05 t-ishii Exp $
+ * $Header$
  *
- * pgpool: a language independent connection pool server for PostgreSQL 
+ * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2006	PgPool Global Development Group
+ * Copyright (c) 2003-2010	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -27,9 +27,14 @@
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <sys/un.h>
+#include <arpa/inet.h>
+#include <netdb.h>
 #ifdef HAVE_NETINET_TCP_H
 #include <netinet/tcp.h>
 #endif
+#ifdef HAVE_SYS_SELECT_H
+#include <sys/select.h>
+#endif
 
 #include <signal.h>
 
@@ -44,25 +49,27 @@
 #include <fcntl.h>
 #endif
 
-#include "pool.h"
-#include "md5.h"
-
-#ifdef NONE_BLOCK
-static void set_nonblock(int fd);
-static void unset_nonblock(int fd);
+#ifdef HAVE_CRYPT_H
+#include <crypt.h>
 #endif
 
-void child_exit(int code);
+#include "pool.h"
+#include "pool_ip.h"
+#include "md5.h"
 
 static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *timeout);
 static StartupPacket *read_startup_packet(POOL_CONNECTION *cp);
 static POOL_CONNECTION_POOL *connect_backend(StartupPacket *sp, POOL_CONNECTION *frontend);
-static void cancel_request(CancelPacket *sp, int secondary_backend);
 static RETSIGTYPE die(int sig);
 static RETSIGTYPE close_idle_connection(int sig);
+static RETSIGTYPE wakeup_handler(int sig);
+static RETSIGTYPE reload_config_handler(int sig);
+static RETSIGTYPE authentication_timeout(int sig);
 static int send_params(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
 static void send_frontend_exits(void);
 static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password);
+static void connection_count_up(void);
+static void connection_count_down(void);
 
 /*
  * non 0 means SIGTERM(smart shutdown) or SIGINT(fast shutdown) has arrived
@@ -70,7 +77,16 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password);
 static int exit_request;
 
 static int idle;		/* non 0 means this child is in idle state */
+static int accepted = 0;
 
+extern int myargc;
+extern char **myargv;
+
+char remote_ps_data[NI_MAXHOST];		/* used for set_ps_display */
+
+volatile sig_atomic_t got_sighup = 0;
+
+int LocalSessionId;	/* Local session id */
 
 /*
 * child main loop
@@ -85,28 +101,31 @@ void do_child(int unix_fd, int inet_fd)
 	struct timeval timeout;
 	static int connected;
 	int connections_count = 0;	/* used if child_max_connections > 0 */
-	int first_ready_for_query_received;		/* for master/slave mode */
 	int found;
+	char psbuf[NI_MAXHOST + 128];
 
 	pool_debug("I am %d", getpid());
 
+	/* Identify myself via ps */
+	init_ps_display("", "", "", "");
+
 	/* set up signal handlers */
 	signal(SIGALRM, SIG_DFL);
 	signal(SIGTERM, die);
 	signal(SIGINT, die);
-	signal(SIGHUP, close_idle_connection);
+	signal(SIGHUP, reload_config_handler);
 	signal(SIGQUIT, die);
 	signal(SIGCHLD, SIG_DFL);
-	signal(SIGUSR1, SIG_DFL);
-	signal(SIGUSR2, SIG_DFL);
+	signal(SIGUSR1, close_idle_connection);
+	signal(SIGUSR2, wakeup_handler);
 	signal(SIGPIPE, SIG_IGN);
 
 #ifdef NONE_BLOCK
 	/* set listen fds to none block */
-	set_nonblock(unix_fd);
+	pool_set_nonblock(unix_fd);
 	if (inet_fd)
 	{
-		set_nonblock(inet_fd);
+		pool_set_nonblock(inet_fd);
 	}
 #endif
 
@@ -161,6 +180,7 @@ void do_child(int unix_fd, int inet_fd)
 		}
 
 		idle = 1;
+		accepted = 0;
 
 		/* perform accept() */
 		frontend = do_accept(unix_fd, inet_fd, &timeout);
@@ -172,25 +192,28 @@ void do_child(int unix_fd, int inet_fd)
 				timeout.tv_sec == 0 && timeout.tv_usec == 0)
 			{
 				pool_debug("child life %d seconds expired", pool_config->child_life_time);
-				send_frontend_exits();
+				/*
+				 * Doesn't need to call this. child_exit() calls it.
+				 * send_frontend_exits();
+				 */
 				child_exit(2);
 			}
 			continue;
 		}
 
 		/* set frontend fd to blocking */
-		unset_nonblock(frontend->fd);
-
-#ifdef NOT_USED
-		set_nonblock(frontend->fd);
-#endif
+		pool_unset_nonblock(frontend->fd);
 
 		/* set busy flag and clear child idle timer */
 		idle = 0;
 		child_idle_sec = 0;
 
-		/* disable timeout */
-		pool_disable_timeout();
+		/* check backend timer is expired */
+		if (backend_timer_expired)
+		{
+			pool_backend_timer();
+			backend_timer_expired = 0;
+		}
 
 		/* read the startup packet */
 	retry_startup:
@@ -199,17 +222,18 @@ void do_child(int unix_fd, int inet_fd)
 		{
 			/* failed to read the startup packet. return to the accept() loop */
 			pool_close(frontend);
+			connection_count_down();
 			continue;
 		}
 
 		/* cancel request? */
 		if (sp->major == 1234 && sp->minor == 5678)
 		{
-			cancel_request((CancelPacket *)sp->startup_packet, 0);
-			if (DUAL_MODE)
-				cancel_request((CancelPacket *)sp->startup_packet, 1);
+			cancel_request((CancelPacket *)sp->startup_packet);
+
 			pool_close(frontend);
 			pool_free_startup_packet(sp);
+			connection_count_down();
 			continue;
 		}
 
@@ -222,6 +246,7 @@ void do_child(int unix_fd, int inet_fd)
 			{
 				pool_close(frontend);
 				pool_free_startup_packet(sp);
+				connection_count_down();
 				continue;
 			}
 
@@ -235,6 +260,29 @@ void do_child(int unix_fd, int inet_fd)
 			goto retry_startup;
 		}
 
+		if (pool_config->enable_pool_hba)
+		{
+			/*
+			 * do client authentication.
+			 * Note that ClientAuthentication does not return if frontend
+			 * was rejected; it simply terminates this process.
+			 */
+			frontend->protoVersion = sp->major;
+			frontend->database = strdup(sp->database);
+			if (frontend->database == NULL)
+			{
+				pool_error("do_child: strdup failed: %s\n", strerror(errno));
+				child_exit(1);
+			}
+			frontend->username = strdup(sp->user);
+			if (frontend->username == NULL)
+			{
+				pool_error("do_child: strdup failed: %s\n", strerror(errno));
+				child_exit(1);
+			}
+			ClientAuthentication(frontend);
+		}
+
 		/*
 		 * Ok, negotiaton with frontend has been done. Let's go to the next step.
 		 */
@@ -244,18 +292,16 @@ void do_child(int unix_fd, int inet_fd)
 		 * we need to connect to the backend and send the startup packet.
 		 */
 
-		first_ready_for_query_received = 0;		/* for master/slave mode */
-
 		/* look for existing connection */
 		found = 0;
-		backend = pool_get_cp(sp->user, sp->database, sp->major);
+		backend = pool_get_cp(sp->user, sp->database, sp->major, 1);
 
 		if (backend != NULL)
 		{
 			found = 1;
 
 			/* existing connection associated with same user/database/major found.
-			 * however we should make sure that the startup packet contents identical.
+			 * however we should make sure that the startup packet contents are identical.
 			 * OPTION data and others might be different.
 			 */
 			if (sp->len != MASTER_CONNECTION(backend)->sp->len)
@@ -283,14 +329,10 @@ void do_child(int unix_fd, int inet_fd)
 			connection_reuse = 0;
 
 			if ((backend = connect_backend(sp, frontend)) == NULL)
+			{
+				connection_count_down();
 				continue;
-
-			/* in master/slave mode, the first "ready for query"
-			 * packet should be treated as if we were not in the
-			 * mode
-			 */
-			if (MASTER_SLAVE)
-				first_ready_for_query_received = 1;
+			}
 		}
 
 		else
@@ -317,6 +359,7 @@ void do_child(int unix_fd, int inet_fd)
 			if (pool_do_reauth(frontend, backend))
 			{
 				pool_close(frontend);
+				connection_count_down();
 				continue;
 			}
 
@@ -325,6 +368,7 @@ void do_child(int unix_fd, int inet_fd)
 				if (send_params(frontend, backend))
 				{
 					pool_close(frontend);
+					connection_count_down();
 					continue;
 				}
 			}
@@ -346,23 +390,36 @@ void do_child(int unix_fd, int inet_fd)
 			if (pool_flush(frontend) < 0)
 			{
 				pool_close(frontend);
+				connection_count_down();
 				continue;
 			}
 
 		}
 
-		/* enable query result read timeout if non strict mode */
-		if (pool_config->replication_strict == 0)
-			pool_enable_timeout();
-
 		connected = 1;
+		LocalSessionId++;
+
+ 		/* show ps status */
+		sp = MASTER_CONNECTION(backend)->sp;
+		snprintf(psbuf, sizeof(psbuf), "%s %s %s idle",
+				 sp->user, sp->database, remote_ps_data);
+		set_ps_display(psbuf, false);
+
+		if (MAJOR(backend) == PROTO_MAJOR_V2)
+			TSTATE(backend) = 'I';
+
+		if (pool_config->load_balance_mode)
+		{
+			/* select load balancing node */
+			backend->info->load_balancing_node = select_load_balancing_node();
+		}
 
 		/* query process loop */
 		for (;;)
 		{
 			POOL_STATUS status;
 
-			status = pool_process_query(frontend, backend, 0, first_ready_for_query_received);
+			status = pool_process_query(frontend, backend, 0);
 
 			sp = MASTER_CONNECTION(backend)->sp;
 
@@ -373,13 +430,15 @@ void do_child(int unix_fd, int inet_fd)
 					/*
 					 * do not cache connection if:
 					 * pool_config->connection_cahe == 0 or
-					 * datase name is template0, template1, or regression
+					 * database name is template0, template1, postgres or regression
 					 */
 					if (pool_config->connection_cache == 0 ||
 						!strcmp(sp->database, "template0") ||
 						!strcmp(sp->database, "template1") ||
+						!strcmp(sp->database, "postgres") ||
 						!strcmp(sp->database, "regression"))
 					{
+						reset_variables();
 						pool_close(frontend);
 						pool_send_frontend_exits(backend);
 						pool_discard_cp(sp->user, sp->database, sp->major);
@@ -389,19 +448,23 @@ void do_child(int unix_fd, int inet_fd)
 						POOL_STATUS status1;
 
 						/* send reset request to backend */
-						status1 = pool_process_query(frontend, backend, 1, 0);
+						status1 = pool_process_query(frontend, backend, 1);
 						pool_close(frontend);
 
 						/* if we detect errors on resetting connection, we need to discard
 						 * this connection since it might be in unknown status
 						 */
 						if (status1 != POOL_CONTINUE)
+						{
+							pool_debug("error in resetting connections. discarding connection pools...");
+							pool_send_frontend_exits(backend);
 							pool_discard_cp(sp->user, sp->database, sp->major);
+						}
 						else
 							pool_connection_pool_timer(backend);
 					}
 					break;
-				
+
 				/* error occured. discard backend connection pool
                    and disconnect connection to the frontend */
 				case POOL_ERROR:
@@ -434,6 +497,9 @@ void do_child(int unix_fd, int inet_fd)
 				break;
 		}
 
+		accepted = 0;
+		connection_count_down();
+
 		timeout.tv_sec = pool_config->child_life_time;
 		timeout.tv_usec = 0;
 
@@ -458,11 +524,10 @@ void do_child(int unix_fd, int inet_fd)
  * -------------------------------------------------------------------
  */
 
-#ifdef NONE_BLOCK
 /*
  * set non-block flag
  */
-static void set_nonblock(int fd)
+void pool_set_nonblock(int fd)
 {
 	int var;
 
@@ -479,12 +544,11 @@ static void set_nonblock(int fd)
 		child_exit(1);
 	}
 }
-#endif
 
 /*
  * unset non-block flag
  */
-static void unset_nonblock(int fd)
+void pool_unset_nonblock(int fd)
 {
 	int var;
 
@@ -509,9 +573,9 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 {
     fd_set	readmask;
     int fds;
+	int save_errno;
 
-	struct sockaddr addr;
-	socklen_t addrlen;
+	SockAddr saddr;
 	int fd = 0;
 	int afd;
 	int inet = 0;
@@ -522,7 +586,12 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 	static int cnt;
 #endif
 	struct timeval *timeoutval;
-	struct timeval tv1, tv2, tmback;
+	struct timeval tv1, tv2, tmback = {0, 0};
+
+	char remote_host[NI_MAXHOST];
+	char remote_port[NI_MAXSERV];
+
+	set_ps_display("wait for connection request", false);
 
 	FD_ZERO(&readmask);
 	FD_SET(unix_fd, &readmask);
@@ -546,6 +615,14 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 
 	fds = select(Max(unix_fd, inet_fd)+1, &readmask, NULL, NULL, timeoutval);
 
+	save_errno = errno;
+	/* check backend timer is expired */
+	if (backend_timer_expired)
+	{
+		pool_backend_timer();
+		backend_timer_expired = 0;
+	}
+
 	/*
 	 * following code fragment computes remaining timeout val in a
 	 * portable way. Linux does this automazically but other platforms do not.
@@ -570,7 +647,7 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 				tmback.tv_usec += 1000000;
 				timeout->tv_sec = tmback.tv_sec;
 				timeout->tv_usec = tmback.tv_usec;
-			} 
+			}
 		}
 #ifdef DEBUG
 		pool_log("g:after select = {%d, %d}", tv2.tv_sec, tv2.tv_usec);
@@ -578,6 +655,8 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 #endif
 	}
 
+	errno = save_errno;
+
 	if (fds == -1)
 	{
 		if (errno == EAGAIN || errno == EINTR)
@@ -608,19 +687,41 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 	 * Note that some SysV systems do not work here. For those
 	 * systems, we need some locking mechanism for the fd.
 	 */
-	addrlen = sizeof(addr);
+	memset(&saddr, 0, sizeof(saddr));
+	saddr.salen = sizeof(saddr.addr);
 
 #ifdef ACCEPT_PERFORMANCE
 	gettimeofday(&now1,0);
 #endif
-	afd = accept(fd, &addr, &addrlen);
+
+ retry_accept:
+
+	/* wait if recovery is started */
+	while (*InRecovery == 1)
+	{
+		pause();
+	}
+
+	afd = accept(fd, (struct sockaddr *)&saddr.addr, &saddr.salen);
+
+	save_errno = errno;
+	/* check backend timer is expired */
+	if (backend_timer_expired)
+	{
+		pool_backend_timer();
+		backend_timer_expired = 0;
+	}
+	errno = save_errno;
 	if (afd < 0)
 	{
+		if (errno == EINTR && *InRecovery)
+			goto retry_accept;
+
 		/*
 		 * "Resource temporarily unavailable" (EAGAIN or EWOULDBLOCK)
-		 * can be silently ignored.
+		 * can be silently ignored. And EINTR can be ignored.
 		 */
-		if (errno != EAGAIN && errno != EWOULDBLOCK)
+		if (errno != EAGAIN && errno != EWOULDBLOCK && errno != EINTR)
 			pool_error("accept() failed. reason: %s", strerror(errno));
 		return NULL;
 	}
@@ -634,6 +735,20 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 	}
 #endif
 
+	/* reload config file */
+	if (got_sighup)
+	{
+		pool_get_config(get_config_file_name(), RELOAD_CONFIG);
+		if (pool_config->enable_pool_hba)
+			load_hba(get_hba_file_name());
+		if (pool_config->parallel_mode)
+			pool_memset_system_db_info(system_db_info->info);
+		got_sighup = 0;
+	}
+
+	connection_count_up();
+	accepted = 1;
+
 	if (pool_config->parallel_mode)
 	{
 		/*
@@ -727,6 +842,20 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 
 	pool_debug("I am %d accept fd %d", getpid(), afd);
 
+	pool_getnameinfo_all(&saddr, remote_host, remote_port);
+	snprintf(remote_ps_data, sizeof(remote_ps_data),
+			 remote_port[0] == '\0' ? "%s" : "%s(%s)",
+			 remote_host, remote_port);
+
+	set_ps_display("accept connection", false);
+
+	/* log who is connecting */
+	if (pool_config->log_connections)
+	{
+		pool_log("connection received: host=%s%s%s",
+				 remote_host, remote_port[0] ? " port=" : "", remote_port);
+	}
+
 	/* set NODELAY and KEEPALIVE options if INET connection */
 	if (inet)
 	{
@@ -755,6 +884,12 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 		close(afd);
 		return NULL;
 	}
+
+	/* save ip addres for hba */
+	memcpy(&cp->raddr, &saddr, sizeof(SockAddr));
+	if (cp->raddr.addr.ss_family == 0)
+		cp->raddr.addr.ss_family = AF_UNIX;
+
 	return cp;
 }
 
@@ -776,6 +911,12 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 		return NULL;
 	}
 
+	if (pool_config->authentication_timeout > 0)
+	{
+		pool_signal(SIGALRM, authentication_timeout);
+		alarm(pool_config->authentication_timeout);
+	}
+
 	/* read startup packet length */
 	if (pool_read(cp, &len, sizeof(len)))
 	{
@@ -788,12 +929,20 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 	{
 		pool_error("read_startup_packet: incorrect packet length (%d)", len);
 	}
+	else if (len >= MAX_STARTUP_PACKET_LENGTH)
+	{
+		pool_error("read_startup_packet: invalid startup packet");
+		pool_free_startup_packet(sp);
+		return NULL;
+	}
 
 	sp->startup_packet = calloc(len, 1);
 	if (!sp->startup_packet)
 	{
 		pool_error("read_startup_packet: out of memory");
 		pool_free_startup_packet(sp);
+		alarm(0);
+		pool_signal(SIGALRM, SIG_IGN);
 		return NULL;
 	}
 
@@ -801,6 +950,8 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 	if (pool_read(cp, sp->startup_packet, len))
 	{
 		pool_free_startup_packet(sp);
+		alarm(0);
+		pool_signal(SIGALRM, SIG_IGN);
 		return NULL;
 	}
 
@@ -820,6 +971,8 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 			{
 				pool_error("read_startup_packet: out of memory");
 				pool_free_startup_packet(sp);
+				alarm(0);
+				pool_signal(SIGALRM, SIG_IGN);
 				return NULL;
 			}
 			strncpy(sp->database, sp2->database, SM_DATABASE);
@@ -829,6 +982,8 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 			{
 				pool_error("read_startup_packet: out of memory");
 				pool_free_startup_packet(sp);
+				alarm(0);
+				pool_signal(SIGALRM, SIG_IGN);
 				return NULL;
 			}
 			strncpy(sp->user, sp2->user, SM_USER);
@@ -848,6 +1003,8 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 					{
 						pool_error("read_startup_packet: out of memory");
 						pool_free_startup_packet(sp);
+						alarm(0);
+						pool_signal(SIGALRM, SIG_IGN);
 						return NULL;
 					}
 				}
@@ -859,6 +1016,8 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 					{
 						pool_error("read_startup_packet: out of memory");
 						pool_free_startup_packet(sp);
+						alarm(0);
+						pool_signal(SIGALRM, SIG_IGN);
 						return NULL;
 					}
 				}
@@ -873,6 +1032,8 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 			{
 				pool_error("read_startup_packet: out of memory");
 				pool_free_startup_packet(sp);
+				alarm(0);
+				pool_signal(SIGALRM, SIG_IGN);
 				return NULL;
 			}
 			sp->user = calloc(1, 1);
@@ -880,6 +1041,8 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 			{
 				pool_error("read_startup_packet: out of memory");
 				pool_free_startup_packet(sp);
+				alarm(0);
+				pool_signal(SIGALRM, SIG_IGN);
 				return NULL;
 			}
 			break;
@@ -887,12 +1050,15 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 		default:
 			pool_error("read_startup_packet: invalid major no: %d", sp->major);
 			pool_free_startup_packet(sp);
+			alarm(0);
+			pool_signal(SIGALRM, SIG_IGN);
 			return NULL;
 	}
 
-	pool_debug("Protocol Major: %d Minor: %d database: %s user: %s", 
+	pool_debug("Protocol Major: %d Minor: %d database: %s user: %s",
 			   sp->major, sp->minor, sp->database, sp->user);
-
+	alarm(0);
+	pool_signal(SIGALRM, SIG_IGN);
 	return sp;
 }
 
@@ -904,23 +1070,41 @@ int send_startup_packet(POOL_CONNECTION_POOL_SLOT *cp)
 	int len;
 
 	len = htonl(cp->sp->len + sizeof(len));
-	pool_write(cp->con, &len, sizeof(len)); 
+	pool_write(cp->con, &len, sizeof(len));
 	return pool_write_and_flush(cp->con, cp->sp->startup_packet, cp->sp->len);
 }
 
 /*
  * process cancel request
  */
-static void cancel_request(CancelPacket *sp, int backend_id)
+void cancel_request(CancelPacket *sp)
 {
 	int	len;
 	int fd;
 	POOL_CONNECTION *con;
 	int i;
+	ConnectionInfo *c = NULL;
+	CancelPacket cp;
+	int loop_num = pool_config->num_init_children*pool_config->max_pool*MAX_NUM_BACKENDS;
 
 	pool_debug("Cancel request received");
 
-	for (i=0;i<NUM_BACKENDS;i++)
+	/* look for cancel key from shmem info */
+	for (i=0;i<loop_num;i++)
+	{
+		c = &con_info[i];
+
+		if (c->pid == sp->pid && c->key == sp->key)
+		{
+			pool_debug("found pid:%d key:%d i:%d",c->pid, c->key,i);
+			c = &con_info[i/(pool_config->max_pool * pool_config->max_pool)];
+			break;
+		}
+	}
+	if (i == loop_num)
+		return;	/* invalid key */
+
+	for (i=0;i<NUM_BACKENDS;i++,c++)
 	{
 		if (!VALID_BACKEND(i))
 			continue;
@@ -943,10 +1127,22 @@ static void cancel_request(CancelPacket *sp, int backend_id)
 		len = htonl(sizeof(len) + sizeof(CancelPacket));
 		pool_write(con, &len, sizeof(len));
 
-		if (pool_write_and_flush(con, sp, sizeof(CancelPacket)) < 0)
+		cp.protoVersion = sp->protoVersion;
+		cp.pid = c->pid;
+		cp.key = c->key;
+
+		pool_debug("pid:%d key: %d",cp.pid,cp.key);
+
+		if (pool_write_and_flush(con, &cp, sizeof(CancelPacket)) < 0)
 			pool_error("Could not send cancel request packet for backend %d", i);
 
 		pool_close(con);
+
+		/*
+		 * this is needed to enure that the next DB node executes the
+		 * query supposed to be canceled.
+		 */
+		sleep(1);
 	}
 }
 
@@ -960,7 +1156,7 @@ static POOL_CONNECTION_POOL *connect_backend(StartupPacket *sp, POOL_CONNECTION
 	if (backend == NULL)
 	{
 		pool_send_error_message(frontend, sp->major, "XX000", "connection cache is full", "",
-								"increace max_pool", __FILE__, __LINE__);
+								"increase max_pool", __FILE__, __LINE__);
 		pool_close(frontend);
 		pool_free_startup_packet(sp);
 		return NULL;
@@ -1032,13 +1228,15 @@ static RETSIGTYPE die(int sig)
 			break;
 	}
 
-	send_frontend_exits();
-
+	/*
+	 * child_exit() does this. So we don't need it.
+	 * send_frontend_exits();
+	 */
 	child_exit(0);
 }
 
 /*
- * signal handler for SIGHUP
+ * signal handler for SIGUSR1
  * close all idle connections
  */
 static RETSIGTYPE close_idle_connection(int sig)
@@ -1053,6 +1251,8 @@ static RETSIGTYPE close_idle_connection(int sig)
 	{
 		if (!MASTER_CONNECTION(p))
 			continue;
+ 		if (!MASTER_CONNECTION(p)->sp)
+			continue;
 		if (MASTER_CONNECTION(p)->sp->user == NULL)
 			continue;
 
@@ -1082,8 +1282,19 @@ static RETSIGTYPE close_idle_connection(int sig)
 }
 
 /*
- * send frontend exiting messages to all connections.
- * this is called when child life time expires or child max connections expires.
+ * signal handler for SIGALRM
+ *
+ */
+static RETSIGTYPE authentication_timeout(int sig)
+{
+	pool_log("authentication is timeout");
+	child_exit(1);
+}
+
+/*
+ * send frontend exiting messages to all connections.  this is called
+ * in any case when child process exits, for example failover, child
+ * life time expires or child max connections expires.
  */
 static void send_frontend_exits(void)
 {
@@ -1102,6 +1313,8 @@ static void send_frontend_exits(void)
 	{
 		if (!MASTER_CONNECTION(p))
 			continue;
+		if (!MASTER_CONNECTION(p)->sp)
+			continue;
 		if (MASTER_CONNECTION(p)->sp->user == NULL)
 			continue;
 		pool_send_frontend_exits(p);
@@ -1149,8 +1362,16 @@ void pool_free_startup_packet(StartupPacket *sp)
 	}
 }
 
+/*
+ * Do house keeping works when pgpool child process exits
+ */
 void child_exit(int code)
 {
+	/* count down global connection counter */
+	if (accepted)
+		connection_count_down();
+
+	/* prepare to shutdown connections to system db */
 	if(pool_config->parallel_mode || pool_config->enable_query_cache)
 	{
 		if (system_db_info->pgconn)
@@ -1158,6 +1379,10 @@ void child_exit(int code)
 		if (pool_system_db_connection())
 			pool_close(pool_system_db_connection()->con);
 	}
+
+	/* let backend know now we are exiting */
+	send_frontend_exits();
+
 	exit(code);
 }
 
@@ -1182,13 +1407,13 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	int len, len1;
 	int status;
 
-	cp = malloc(sizeof(POOL_CONNECTION_POOL));
+	cp = malloc(sizeof(POOL_CONNECTION_POOL_SLOT));
 	if (cp == NULL)
 	{
 		pool_error("make_persistent_db_connection: could not allocate memory");
 		return NULL;
 	}
-	memset(cp, 0, sizeof(POOL_CONNECTION_POOL));
+	memset(cp, 0, sizeof(POOL_CONNECTION_POOL_SLOT));
 
 	startup_packet = malloc(sizeof(*startup_packet));
 	if (startup_packet == NULL)
@@ -1368,7 +1593,7 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 		int size;
 		char salt[3];
 		char *crypt_password;
-		
+
 		status = pool_read(cp->con, &salt, 2);
 		if (status > 0)
 		{
@@ -1434,7 +1659,7 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 	}
 	else
 	{
-		pool_error("s_do_auth: auth kind %d not supported yet");
+		pool_error("s_do_auth: auth kind %d not supported yet", auth_kind);
 		return -1;
 	}
 
@@ -1545,3 +1770,86 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 	}
 	return -1;
 }
+
+/*
+ * Count up connection counter (from frontend to pgpool)
+ * in shared memory
+ */
+static void connection_count_up(void)
+{
+	pool_semaphore_lock(CONN_COUNTER_SEM);
+	Req_info->conn_counter++;
+	pool_semaphore_unlock(CONN_COUNTER_SEM);
+}
+
+/*
+ * Count down connection counter (from frontend to pgpool)
+ * in shared memory
+ */
+static void connection_count_down(void)
+{
+	pool_semaphore_lock(CONN_COUNTER_SEM);
+	/*
+	 * Make sure that we do not decrement too much.  If failed to read
+	 * a start up packet, or receive cancel request etc.,
+	 * connection_count_down() is called and goes back to the
+	 * connection accept loop. Problem is, at the very beginning of
+	 * the connection accept loop, if we have received a signal, we
+	 * call child_exit() which calls connection_count_down() again.
+	 */
+	if (Req_info->conn_counter > 0)
+		Req_info->conn_counter--;
+	pool_semaphore_unlock(CONN_COUNTER_SEM);
+}
+
+/*
+ * handle SIGUSR2
+ * Wakeup all process
+ */
+static RETSIGTYPE wakeup_handler(int sig)
+{
+}
+
+
+/*
+ * Select load balancing node
+ */
+int select_load_balancing_node(void)
+{
+	double total_weight,r;
+	int i;
+
+	/* choose a backend in random manner with weight */
+	selected_slot = MASTER_NODE_ID;
+	total_weight = 0.0;
+
+	for (i=0;i<NUM_BACKENDS;i++)
+	{
+		if (VALID_BACKEND(i))
+		{
+			total_weight += BACKEND_INFO(i).backend_weight;
+		}
+	}
+	r = (((double)random())/RAND_MAX) * total_weight;
+	total_weight = 0.0;
+	for (i=0;i<NUM_BACKENDS;i++)
+	{
+		if (VALID_BACKEND(i) && BACKEND_INFO(i).backend_weight > 0.0)
+		{
+			if(r >= total_weight)
+				selected_slot = i;
+			else
+				break;
+			total_weight += BACKEND_INFO(i).backend_weight;
+		}
+	}
+
+	pool_debug("select_load_balancing_node: selected backend id is %d", selected_slot);
+	return selected_slot;
+}
+
+/* SIGHUP handler */
+static RETSIGTYPE reload_config_handler(int sig)
+{
+	got_sighup = 1;
+}
