@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/main.c,v 1.45.2.6 2009/09/25 07:35:16 t-ishii Exp $
+ * $Header$
  *
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
@@ -277,7 +277,7 @@ int main(int argc, char **argv)
 		load_hba(hba_file);
 
 	/*
-	 * if a non-switch argument remains, then it should be either "reload", "stop" or "switch"
+	 * If a non-switch argument remains, then it should be either "reload" or "stop".
 	 */
 	if (optind == (argc - 1))
 	{
@@ -384,7 +384,16 @@ int main(int argc, char **argv)
 		inet_fd = create_inet_domain_socket(pool_config->listen_addresses, pool_config->port);
 	}
 
-	size = pool_config->num_init_children * pool_config->max_pool * sizeof(ConnectionInfo);
+	/*
+	 * con_info is a 3 dimension array: i corresponds to pgpool child
+	 * process, j corresponds to connection pool in each process and k
+	 * corresponds to backends in each connection pool.
+	 *
+	 * XXX: Before 2010/4/12 this was a 2 dimension array: i
+	 * corresponds to pgpool child process, j corresponds to
+	 * connection pool in each process. Of course this was wrong.
+	 */
+	size = pool_config->num_init_children * pool_config->max_pool * MAX_NUM_BACKENDS * sizeof(ConnectionInfo);
 	con_info = pool_shared_memory_create(size);
 	if (con_info == NULL)
 	{
@@ -403,7 +412,7 @@ int main(int argc, char **argv)
 	memset(pids, 0, size);
 	for (i = 0; i < pool_config->num_init_children; i++)
 	{
-		pids[i].connection_info = &con_info[i * pool_config->max_pool];
+		pids[i].connection_info = &con_info[i * pool_config->max_pool * MAX_NUM_BACKENDS];
 	}
 
 	/* create fail over/switch over event area */
@@ -853,6 +862,7 @@ static int create_inet_domain_socket(const char *hostname, const int port)
 	int status;
 	int one = 1;
 	int len;
+	int backlog;
 
 	fd = socket(AF_INET, SOCK_STREAM, 0);
 	if (fd == -1)
@@ -902,7 +912,11 @@ static int create_inet_domain_socket(const char *hostname, const int port)
 		myexit(1);
 	}
 
-	status = listen(fd, PGPOOLMAXLITSENQUEUELENGTH);
+	backlog = pool_config->num_init_children * 2;
+	if (backlog > PGPOOLMAXLITSENQUEUELENGTH)
+		backlog = PGPOOLMAXLITSENQUEUELENGTH;
+
+	status = listen(fd, backlog);
 	if (status < 0)
 	{
 		pool_error("listen() failed. reason: %s", strerror(errno));
@@ -1570,7 +1584,7 @@ static RETSIGTYPE reap_handler(int sig)
 
 /*
  * Attach zombie processes and restart child processes.
- * reaper() must be called under protecting signals.
+ * reaper() must be called protected from signals.
  */
 static void reaper(void)
 {
@@ -1603,16 +1617,29 @@ static void reaper(void)
 	while ((pid = wait3(&status, WNOHANG, NULL)) > 0)
 #endif
 	{
+		if (WIFSIGNALED(status) && WTERMSIG(status) == SIGSEGV)
+		{
+			/* Child terminated by segmentation fault. Report it */
+			pool_error("Child process %d was terminated by segmentation fault", pid);
+		}
+			
 		/* if exiting child process was PCP handler */
 		if (pid == pcp_pid)
 		{
-			pool_debug("PCP child %d exits with status %d by signal %d", pid, status, WTERMSIG(status));
+			if (WIFSIGNALED(status))
+				pool_debug("PCP child %d exits with status %d by signal %d", pid, status, WTERMSIG(status));
+			else
+				pool_debug("PCP child %d exits with status %d", pid, status);
 
 			pcp_pid = pcp_fork_a_child(pcp_unix_fd, pcp_inet_fd, pcp_conf_file);
 			pool_debug("fork a new PCP child pid %d", pcp_pid);
 			break;
-		} else {
-			pool_debug("child %d exits with status %d by signal %d", pid, status, WTERMSIG(status));
+		} else
+		{
+			if (WIFSIGNALED(status))
+				pool_debug("child %d exits with status %d by signal %d", pid, status, WTERMSIG(status));
+			else
+				pool_debug("child %d exits with status %d", pid, status);
 
 			/* look for exiting child's pid */
 			for (i=0;i<pool_config->num_init_children;i++)
