@@ -1,5 +1,21 @@
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/pcp/pcp_proc_info.c,v 1.1.1.1 2006/09/08 03:36:50 t-ishii Exp $
+ * $Header$
+ *
+ * pgpool: a language independent connection pool server for PostgreSQL 
+ * written by Tatsuo Ishii
+ *
+ * Copyright (c) 2003-2008	PgPool Global Development Group
+ *
+ * Permission to use, copy, modify, and distribute this software and
+ * its documentation for any purpose and without fee is hereby
+ * granted, provided that the above copyright notice appear in all
+ * copies and that both that copyright notice and this permission
+ * notice appear in supporting documentation, and that the name of the
+ * author not be used in advertising or publicity pertaining to
+ * distribution of the software without specific, written prior
+ * permission. The author makes no representations about the
+ * suitability of this software for any purpose.  It is provided "as
+ * is" without express or implied warranty.
  *
  * Client program to send "process info" command.
  */
@@ -7,6 +23,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <unistd.h>
 
 #include "pcp.h"
 
@@ -24,36 +41,47 @@ main(int argc, char **argv)
 	int processID;
 	ProcessInfo *process_info;
 	int array_size;
+	int ch;
 
-	if (argc == 2 && (strcmp(argv[1], "-h") == 0) )
-	{
-		usage();
-		exit(0);
+	while ((ch = getopt(argc, argv, "hd")) != -1) {
+		switch (ch) {
+		case 'd':
+			pcp_enable_debug();
+			break;
+
+		case 'h':
+		case '?':
+		default:
+			usage();
+			exit(0);
+		}
 	}
+	argc -= optind;
+	argv += optind;
 
-	if (argc != 7)
+	if (argc != 6)
 	{
 		errorcode = INVALERR;
 		pcp_errorstr(errorcode);
 		myexit(errorcode);
 	}
 
-	timeout = atol(argv[1]);
+	timeout = atol(argv[0]);
 	if (timeout < 0) {
 		errorcode = INVALERR;
 		pcp_errorstr(errorcode);
 		myexit(errorcode);
 	}
 
-	if (strlen(argv[2]) >= MAX_DB_HOST_NAMELEN)
+	if (strlen(argv[1]) >= MAX_DB_HOST_NAMELEN)
 	{
 		errorcode = INVALERR;
 		pcp_errorstr(errorcode);
 		myexit(errorcode);
 	}
-	strcpy(host, argv[2]);
+	strcpy(host, argv[1]);
 
-	port = atoi(argv[3]);
+	port = atoi(argv[2]);
 	if (port <= 1024 || port > 65535)
 	{
 		errorcode = INVALERR;
@@ -61,23 +89,23 @@ main(int argc, char **argv)
 		myexit(errorcode);
 	}
 
+	if (strlen(argv[3]) >= MAX_USER_PASSWD_LEN)
+	{
+		errorcode = INVALERR;
+		pcp_errorstr(errorcode);
+		myexit(errorcode);
+	}
+	strcpy(user, argv[3]);
+
 	if (strlen(argv[4]) >= MAX_USER_PASSWD_LEN)
 	{
 		errorcode = INVALERR;
 		pcp_errorstr(errorcode);
 		myexit(errorcode);
 	}
-	strcpy(user, argv[4]);
+	strcpy(pass, argv[4]);
 
-	if (strlen(argv[5]) >= MAX_USER_PASSWD_LEN)
-	{
-		errorcode = INVALERR;
-		pcp_errorstr(errorcode);
-		myexit(errorcode);
-	}
-	strcpy(pass, argv[5]);
-
-	processID = atoi(argv[6]);
+	processID = atoi(argv[5]);
 	if (processID < 0)
 	{
 		errorcode = INVALERR;
@@ -126,8 +154,9 @@ static void
 usage(void)
 {
 	fprintf(stderr, "pcp_proc_info - display a pgpool-II child process' information\n\n");
-	fprintf(stderr, "Usage: pcp_proc_info timeout hostname port# username password PID\n");
+	fprintf(stderr, "Usage: pcp_proc_info [-d] timeout hostname port# username password PID\n");
 	fprintf(stderr, "Usage: pcp_proc_info -h\n\n");
+	fprintf(stderr, "  -d       - enable debug message (optional)\n");
 	fprintf(stderr, "  timeout  - connection timeout value in seconds. command exits on timeout\n");
 	fprintf(stderr, "  hostname - pgpool-II hostname\n");
 	fprintf(stderr, "  port#    - pgpool-II port number\n");
