@@ -1,12 +1,12 @@
 /* -*-pgsql-c-*- */
 /*
  *
- * $Header: /cvsroot/pgpool/pgpool-II/pool.h,v 1.5.2.5 2007/07/26 09:03:04 y-asaba Exp $
+ * $Header$
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2007	PgPool Global Development Group
+ * Copyright (c) 2003-2010	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -33,6 +33,7 @@
 #include <stdio.h>
 #include <time.h>
 #include <sys/types.h>
+#include <limits.h>
 
 /* undef this if you have problems with non blocking accept() */
 #define NONE_BLOCK
@@ -55,13 +56,7 @@
 #define DEFAULT_SOCKET_DIR "/tmp"
 
 /* pid file name */
-#define PID_FILE_NAME "pgpool.pid"
-
-/* strict mode comment in SQL */
-#define STRICT_MODE_STR "/*STRICT*/"
-#define STRICT_MODE(s) (strncasecmp((s), STRICT_MODE_STR, strlen(STRICT_MODE_STR)) == 0)
-#define NO_STRICT_MODE_STR "/*NO STRICT*/"
-#define NO_STRICT_MODE(s) (strncasecmp((s), NO_STRICT_MODE_STR, strlen(NO_STRICT_MODE_STR)) == 0)
+#define DEFAULT_PID_FILE_NAME "/var/run/pgpool/pgpool.pid"
 
 typedef enum {
 	POOL_CONTINUE = 0,
@@ -72,10 +67,29 @@ typedef enum {
 	POOL_DEADLOCK
 } POOL_STATUS;
 
+
+typedef enum {
+	INIT_CONFIG = 1,   /* 0x01 */
+	RELOAD_CONFIG = 2  /* 0x02 */
+} POOL_CONFIG_CONTEXT;
+
+
 /* protocol major version numbers */
 #define PROTO_MAJOR_V2	2
 #define PROTO_MAJOR_V3	3
 
+/* Cancel packet proto major */
+#define PROTO_CANCEL	80877102
+
+/*
+ * In protocol 3.0 and later, the startup packet length is not fixed, but
+ * we set an arbitrary limit on it anyway.	This is just to prevent simple
+ * denial-of-service attacks via sending enough data to run the server
+ * out of memory.
+ */
+#define MAX_STARTUP_PACKET_LENGTH 10000
+
+
 typedef struct StartupPacket_v2
 {
 	int			protoVersion;		/* Protocol version */
@@ -118,24 +132,19 @@ typedef struct {
     int	child_life_time;	/* if idle for this seconds, child exits */
     int	connection_life_time;	/* if idle for this seconds, connection closes */
     int	child_max_connections;	/* if max_connections received, child exits */
+	int client_idle_limit;		/* If client_idle_limit is n (n > 0), the client is forced to be
+								   disconnected after n seconds idle */
+	int authentication_timeout; /* maximum time in seconds to complete client authentication */
     int	max_pool;	/* max # of connection pool per child */
     char *logdir;		/* logging directory */
+    char *pid_file_name;		/* pid file name */
     char *backend_socket_dir;	/* Unix domain socket directory for the PostgreSQL server */
 	int replication_mode;		/* replication mode */
-	int replication_strict;	/* if non 0, wait for completion of the
-                               query sent to master to avoid deadlock */
 
 	int log_connections;		/* 0:false, 1:true - logs incoming connections */
 	int log_hostname;		/* 0:false, 1:true - resolve hostname */
 	int enable_pool_hba;		/* 0:false, 1:true - enables pool_hba.conf file authentication */
 
-	/*
-	 * if next PostgreSQL server does not respond in this milli
-	 * seconds, abort this session.  this is not compatible with
-	 * replication_strict = 1. 0 means no timeout.
-	 */
-	int replication_timeout;
-
 	int load_balance_mode;		/* load balance mode */
 
 	int replication_stop_on_mismatch;		/* if there's a data mismatch between master and secondary
@@ -150,6 +159,16 @@ typedef struct {
 	int health_check_timeout;	/* health check timeout */
 	int health_check_period;	/* health check period */
 	char *health_check_user;		/* PostgreSQL user name for health check */
+	char *failover_command;     /* execute command when failover happens */
+	char *failback_command;     /* execute command when failback happens */
+	char *recovery_user;		/* PostgreSQL user name for online recovery */
+	char *recovery_password;		/* PostgreSQL user password for online recovery */
+	char *recovery_1st_stage_command;   /* Online recovery command in 1st stage */
+	char *recovery_2nd_stage_command;   /* Online recovery command in 2nd stage */
+	int recovery_timeout;				/* maximum time in seconds to wait for remote start-up */
+	int client_idle_limit_in_recovery;		/* If > 0, the client is forced to be
+											 *  disconnected after n seconds idle
+											 *  This parameter is only valid while in recovery 2nd statge */
 	int insert_lock;	/* if non 0, automatically lock table with INSERT to keep SERIAL
 						   data consistency */
 	int ignore_leading_white_space;		/* ignore leading white spaces of each query */
@@ -255,7 +274,7 @@ typedef struct {
 } POOL_CONNECTION_POOL_SLOT;
 
 typedef struct {
-	ConnectionInfo *info;
+	ConnectionInfo *info;		/* connection info on shmem */
     POOL_CONNECTION_POOL_SLOT	*slots[MAX_NUM_BACKENDS];
 } POOL_CONNECTION_POOL;
 
@@ -290,20 +309,58 @@ typedef struct {
 	UNIT unit;
 } Interval;
 
+/*
+ * Relation cache structure
+ */
+#define MAX_ITEM_LENGTH	1024
+
+/* Relation lookup cache structure */
+
+typedef void *(*func_ptr) ();
+
+typedef struct {
+	char dbname[MAX_ITEM_LENGTH];	/* database name */
+	char relname[MAX_ITEM_LENGTH];	/* table name */
+	void *data;	/* user data */
+	int refcnt;		/* reference count */
+	int session_id;		/* LocalSessionId */
+} PoolRelCache;
+
+typedef struct {
+	int num;		/* number of cache items */
+	char sql[MAX_ITEM_LENGTH];	/* Query to relation */
+	/*
+	 * User defined function to be called at data register.
+	 * Argument is POOL_SELECT_RESULT *.
+	 * This function must return a pointer to be
+	 * saved in cache->data.
+	 */
+	func_ptr	register_func;
+	/*
+	 * User defined function to be called at data unregister.
+	 * Argument cache->data.
+	 */
+	func_ptr	unregister_func;
+	bool cache_is_session_local;		/* True if cache life time is session local */
+	PoolRelCache *cache;	/* cache data */
+} POOL_RELCACHE;
+
+
 #ifdef NOT_USED
 #define NUM_BACKENDS (in_load_balance? (selected_slot+1) : \
 					  (((!REPLICATION && !PARALLEL_MODE)||master_slave_dml)? Req_info->master_node_id+1: \
 					   pool_config->backend_desc->num_backends))
 #endif
-#define NUM_BACKENDS (in_load_balance? (selected_slot+1) : \
-					  (((!REPLICATION && !PARALLEL_MODE) && !MASTER_SLAVE) ? Req_info->master_node_id+1: \
-					   pool_config->backend_desc->num_backends))
+/* NUM_BACKENDS now always returns actual number of backends if not in_load_balance */
+#define NUM_BACKENDS (in_load_balance ? (selected_slot+1) : pool_config->backend_desc->num_backends)
 #define BACKEND_INFO(backend_id) (pool_config->backend_desc->backend_info[(backend_id)])
 #define LOAD_BALANCE_STATUS(backend_id) (pool_config->load_balance_status[(backend_id)])
+/* if RAW_MODE, VALID_BACKEND returns the selected node only */
 #define VALID_BACKEND(backend_id) \
+	(RAW_MODE ? (backend_id) == MASTER_NODE_ID : \
 	(in_load_balance ? LOAD_BALANCE_STATUS(backend_id) == LOAD_SELECTED : \
     ((BACKEND_INFO(backend_id).backend_status == CON_UP) || \
-	 (BACKEND_INFO(backend_id).backend_status == CON_CONNECT_WAIT)))
+	 (BACKEND_INFO(backend_id).backend_status == CON_CONNECT_WAIT))))
 #define CONNECTION_SLOT(p, slot) ((p)->slots[(slot)])
 #define CONNECTION(p, slot) (CONNECTION_SLOT(p, slot)->con)
 #define MASTER_CONNECTION(p) ((p)->slots[MASTER_NODE_ID])
@@ -314,6 +371,7 @@ typedef struct {
 #define MASTER_SLAVE (pool_config->master_slave_enabled)
 #define DUAL_MODE (REPLICATION || MASTER_SLAVE)
 #define PARALLEL_MODE (pool_config->parallel_mode)
+#define RAW_MODE (!REPLICATION && !PARALLEL_MODE && !MASTER_SLAVE)
 #define MASTER(p) MASTER_CONNECTION(p)->con
 //#define SECONDARY(p) SECONDARY_CONNECTION(p)->con
 #define MAJOR(p) MASTER_CONNECTION(p)->sp->major
@@ -329,8 +387,12 @@ typedef struct {
 #define LOCK_COMMENT_SZ (sizeof(LOCK_COMMENT)-1)
 #define NO_LOCK_COMMENT "/*NO INSERT LOCK*/"
 #define NO_LOCK_COMMENT_SZ (sizeof(NO_LOCK_COMMENT)-1)
+#define NO_LOAD_BALANCE "/*NO LOAD BALANCE*/"
+#define NO_LOAD_BALANCE_COMMENT_SZ (sizeof(NO_LOAD_BALANCE)-1)
 
 #define MAX_NUM_SEMAPHORES		3
+#define CONN_COUNTER_SEM 0
+#define REQUEST_INFO_SEM 1
 
 #define MY_PROCESS_INFO (pids[my_proc_id])
 
@@ -349,45 +411,91 @@ typedef enum SemNum
  */
 typedef enum {
 	NODE_UP_REQUEST = 0,
-	NODE_DOWN_REQUEST
+	NODE_DOWN_REQUEST,
+	NODE_RECOVERY_REQUEST,
+	CLOSE_IDLE_REQUEST
 } POOL_REQUEST_KIND;
 
 typedef struct {
 	POOL_REQUEST_KIND	kind;	/* request kind */
-	int node_id;		/* request node id */
+	int node_id[MAX_NUM_BACKENDS];		/* request node id */
 	int master_node_id;	/* the youngest node id which is not in down status */
+	int conn_counter;
 } POOL_REQUEST_INFO;
 
+/* description of row. corresponding to RowDescription message */
+typedef struct {
+	char *attrname;		/* attribute name */
+	int oid;	/* 0 or non 0 if it's a table oblect */
+	int attrnumber;		/* attribute number starting with 1. 0 if it's not a table */
+	int typeoid;		/* data type oid */
+	int	size;	/* data length minus means variable data type */
+	int mod;	/* data type modifier */
+} AttrInfo;
+
+typedef struct {
+	int num_attrs;		/* number of attributes */
+	AttrInfo *attrinfo;
+} RowDesc;
+
+typedef struct {
+	RowDesc *rowdesc;	/* attribute info */
+	int numrows;		/* number of rows */
+	int *nullflags;	/* if NULL, -1 or length of the string excluding termination null */
+	char **data;		/* actual row character data terminated with null */
+} POOL_SELECT_RESULT;
+
 /*
  * global variables
  */
+extern pid_t mypid; /* parent pid */
 extern POOL_CONFIG *pool_config;	/* configuration values */
 extern POOL_CONNECTION_POOL *pool_connection_pool;	/* connection pool */
 extern volatile sig_atomic_t backend_timer_expired; /* flag for connection closed timer is expired */
 extern long int weight_master;	/* normalized weight of master (0-RAND_MAX range) */
 extern int my_proc_id;  /* process table id (!= UNIX's PID) */
 extern POOL_SYSTEMDB_CONNECTION_POOL *system_db_info; /* systemdb */
-extern ProcessInfo *pids; /* process information table */
+extern ProcessInfo *pids; /* shmem process information table */
+extern ConnectionInfo *con_info; /* shmem connection info table */
 extern int in_load_balance;		/* non 0 if in load balance mode */
 extern int selected_slot;		/* selected DB node for load balance */
 extern int master_slave_dml;	/* non 0 if master/slave mode is specified in config file */
 extern POOL_REQUEST_INFO *Req_info;
+extern volatile sig_atomic_t *InRecovery;
 extern char remote_ps_data[];		/* used for set_ps_display */
+extern volatile sig_atomic_t got_sighup;
+
+#define QUERY_STRING_BUFFER_LEN 1024
+extern char query_string_buffer[];		/* last query string sent to simpleQuery() */
+
+extern int LocalSessionId;	/* Local session id. incremented when new frontend connected */
 
 /*
  * public functions
  */
+extern char *get_config_file_name(void);
+extern char *get_hba_file_name(void);
+#ifdef __GNUC__
+extern void pool_error(const char *fmt,...)
+   	__attribute__((format (printf, 1, 2)));
+extern void pool_debug(const char *fmt,...)
+   	__attribute__((format (printf, 1, 2)));
+extern void pool_log(const char *fmt,...)
+   	__attribute__((format (printf, 1, 2)));
+#else
 extern void pool_error(const char *fmt,...);
 extern void pool_debug(const char *fmt,...);
 extern void pool_log(const char *fmt,...);
-extern int pool_get_config(char *confpath);
+#endif
+extern int pool_init_config(void);
+extern int pool_get_config(char *confpath, POOL_CONFIG_CONTEXT context);
 extern void do_child(int unix_fd, int inet_fd);
 extern void pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file);
+extern int select_load_balancing_node(void);
 extern int pool_init_cp(void);
 extern POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 									  POOL_CONNECTION_POOL *backend,
-									  int connection_reuse,
-									  int first_ready_for_query_received);
+									  int reset_request);
 
 extern POOL_CONNECTION *pool_open(int fd);
 extern void pool_close(POOL_CONNECTION *cp);
@@ -416,6 +524,7 @@ extern POOL_STATUS NoticeResponse(POOL_CONNECTION *frontend,
 								  POOL_CONNECTION_POOL *backend);
 
 extern void notice_backend_error(int node_id);
+extern void degenerate_backend_set(int *node_id_set, int count);
 extern void send_failback_request(int node_id);
 
 extern void pool_connection_pool_timer(POOL_CONNECTION_POOL *backend);
@@ -426,9 +535,8 @@ extern int connect_unix_domain_socket(int secondary_backend);
 extern int connect_inet_domain_socket_by_port(char *host, int port);
 extern int connect_unix_domain_socket_by_port(int port, char *socket_dir);
 
-extern int pool_check_fd(POOL_CONNECTION *cp, int notimeout);
-extern void pool_enable_timeout(void);
-extern void pool_disable_timeout(void);
+extern void pool_set_timeout(int timeoutval);
+extern int pool_check_fd(POOL_CONNECTION *cp);
 
 extern void pool_send_frontend_exits(POOL_CONNECTION_POOL *backend);
 
@@ -455,6 +563,21 @@ extern void pool_send_error_message(POOL_CONNECTION *frontend, int protoMajor,
 							 char *hint,
 							 char *file,
 							 int line);
+extern void pool_send_fatal_message(POOL_CONNECTION *frontend, int protoMajor,
+							 char *code,
+							 char *message,
+							 char *detail,
+							 char *hint,
+							 char *file,
+							 int line);
+extern void pool_send_severity_message(POOL_CONNECTION *frontend, int protoMajor,
+							 char *code,
+							 char *message,
+							 char *detail,
+							 char *hint,
+							 char *file,
+							 char *severity,
+							 int line);
 extern void pool_send_readyforquery(POOL_CONNECTION *frontend);
 extern int send_startup_packet(POOL_CONNECTION_POOL_SLOT *cp);
 extern void pool_free_startup_packet(StartupPacket *sp);
@@ -482,9 +605,12 @@ extern POOL_STATUS OneNode_do_command(POOL_CONNECTION *frontend, POOL_CONNECTION
 extern POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	char *hostname, int port, char *dbname, char *user, char *password);
 
+extern POOL_STATUS do_query(POOL_CONNECTION *backend, char *query, POOL_SELECT_RESULT **result);
+
 /* define pool_system.c */
 extern POOL_CONNECTION_POOL_SLOT *pool_system_db_connection(void);
 extern DistDefInfo *pool_get_dist_def_info (char * dbname, char * schema_name, char * table_name);
+extern RepliDefInfo *pool_get_repli_def_info (char * dbname, char * schema_name, char * table_name);
 extern int pool_get_id (DistDefInfo *info, const char * value);
 extern int system_db_connect (void);
 extern int pool_memset_system_db_info (SystemDBInfo *info);
@@ -514,4 +640,27 @@ extern void init_ps_display(const char *username, const char *dbname,
 extern void set_ps_display(const char *activity, bool force);
 extern const char *get_ps_display(int *displen);
 
+/* recovery.c */
+extern int start_recovery(int recovery_node);
+extern void finish_recovery(void);
+
+/* child.c */
+extern void pool_set_nonblock(int fd);
+extern void pool_unset_nonblock(int fd);
+extern void cancel_request(CancelPacket *sp);
+
+/* pool_process_query.c */
+void free_select_result(POOL_SELECT_RESULT *result);
+void reset_variables(void);
+POOL_STATUS pool_extract_error_message(POOL_CONNECTION *backend, int major, bool unread, char **message);
+
+/* pool_relcache.c */
+POOL_RELCACHE *pool_create_relcache(int cachesize, char *sql,
+									func_ptr register_func, func_ptr unregister_func,
+									bool issessionlocal);
+void pool_discard_relcache(POOL_RELCACHE *relcache);
+void *pool_search_relcache(POOL_RELCACHE *relcache, POOL_CONNECTION_POOL *backend, char *table);
+void *int_register_func(POOL_SELECT_RESULT *res);
+void *int_unregister_func(void *data);
+
 #endif /* POOL_H */
